// Global objects/variables
//-----------------------------------------
SamplerState SamPointClamp
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState SamLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

RasterizerState CullNone
{
    CullMode = None;
};

BlendState AlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = ONE;
    DestBlendAlpha = ONE;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
}; 

BlendState AlphaBlendingColor
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_COLOR;
    DestBlend = INV_SRC_COLOR;
    BlendOp = ADD;
    SrcBlendAlpha = ONE;
    DestBlendAlpha = ONE;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
}; 

BlendState NoAlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
    RenderTargetWriteMask[0] = 0x0F;
};

// Textures
Texture3D TextureVelocity0;
Texture3D TextureVelocity1;
Texture3D TextureColor0;
Texture3D TextureColor1;
Texture3D TextureTempScalar0;
Texture3D TextureTempScalar1;
Texture3D TexturePressure;
Texture3D TextureObstacles;

// Variables
float TextureWidth, TextureHeight, TextureDepth;
float Timestep;
float3 BoxMin, BoxMax;
float4 BoxColor;

// Helper Functions
//-----------------------------------------

float3 ToTexCoord(float3 x)
{
	return float3(x.x / TextureWidth, x.y / TextureHeight, (x.z + 0.5f) / TextureDepth);
}

bool IsObstacle(float3 texCoords)
{
    return TextureObstacles.SampleLevel(SamPointClamp, texCoords, 0).r > 0.01f;
}

bool IsWater(float3 texCoords)
{
	float2 val = TextureColor0.SampleLevel(SamPointClamp, texCoords, 0).ra;
	return  val.r > -0.1 || val.g < 0;
}

bool IsWaterStrict(float3 texCoords)
{
	return TextureColor0.SampleLevel(SamPointClamp, texCoords, 0).a < 1;
}

// Global Vertex/GeometryShader
//-----------------------------------------

struct VS_OUTPUT
{
    float4 Pos : SV_POSITION;
    float3 Tex : TEXCOORD0;	// 3D cell texcoords
};

struct GS_OUTPUT
{
    float4 Pos               : SV_Position; 
    float3 Tex				 : TEXCOORD0;   
    uint RTIndex             : SV_RenderTargetArrayIndex;  // Destination slice
};

VS_OUTPUT VS(float4 pos : POSITION)
{
    VS_OUTPUT output;
    output.Pos.xy = pos.xy;
    output.Pos.z = 0;
    output.Pos.w = 1;
    
    output.Tex.x = (int)((pos.x / 2.0f + 0.5f) * TextureWidth);
	output.Tex.y = (int)((1.0f - (pos.y / 2.0f + 0.5f)) * TextureHeight);
    output.Tex.z = (int)(pos.z * TextureDepth);
    
    return output;
}

[maxvertexcount(3)]
void GS(triangle VS_OUTPUT input[3], inout TriangleStream<GS_OUTPUT> triStream)
{
    GS_OUTPUT output;
    
    output.RTIndex = input[0].Tex.z;
    
    for(int v = 0; v < 3; v++)
    {
        output.Pos = input[v].Pos; 
        output.Tex = input[v].Tex;
        triStream.Append(output);
    }
    
    triStream.RestartStrip();
}

// Velocity advection
//-----------------------------------------

float4 PSAdvectVelocity(GS_OUTPUT input) : SV_Target
{	
	if(!IsWater(ToTexCoord(input.Tex)))
		return 0;
	float3 pos = input.Tex;
	
    pos -= Timestep * TextureVelocity1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0).xyz;

	float4 vel = TextureVelocity1.SampleLevel(SamLinear, ToTexCoord(pos), 0);
	
	vel.y -= 25 * Timestep;		// Gravity

	return vel;
}

// Color advection
//-----------------------------------------

float4 PSAdvectColor(GS_OUTPUT input) : SV_Target
{
    if(IsObstacle(ToTexCoord(input.Tex)))
		return TextureColor0.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);

    float3 pos = input.Tex;
	
	pos -= Timestep * TextureVelocity1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0).xyz;
	
	return TextureColor0.SampleLevel(SamLinear, ToTexCoord(pos), 0);
}

// Divergence computation
//-----------------------------------------

float PSComputeDivergence(GS_OUTPUT input) : SV_Target
{    
    if(!IsWaterStrict(ToTexCoord(input.Tex)))
		return 0;
	
	float3 center = TextureVelocity1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);

	float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float3 left   = TextureVelocity1.SampleLevel(SamPointClamp, leftAddr, 0);
    float3 right  = TextureVelocity1.SampleLevel(SamPointClamp, rightAddr, 0);
    float3 bottom = TextureVelocity1.SampleLevel(SamPointClamp, bottomAddr, 0);
    float3 top    = TextureVelocity1.SampleLevel(SamPointClamp, topAddr, 0);
    float3 down   = TextureVelocity1.SampleLevel(SamPointClamp, downAddr, 0);
    float3 up     = TextureVelocity1.SampleLevel(SamPointClamp, upAddr, 0);

	if(!IsWaterStrict(leftAddr))     left = 0;
    if(!IsWaterStrict(rightAddr))   right = 0;
    if(!IsWaterStrict(bottomAddr)) bottom = 0;
    if(!IsWaterStrict(topAddr))       top = 0;
    if(!IsWaterStrict(downAddr))     down = 0;
    if(!IsWaterStrict(upAddr))         up = 0;

	if(IsObstacle(leftAddr) )     left = 0;
    if(IsObstacle(rightAddr) )   right = 0;
    if(IsObstacle(bottomAddr))  bottom = 0;
    if(IsObstacle(topAddr)  )      top = 0;
    if(IsObstacle(downAddr) )     down = 0;
    if(IsObstacle(upAddr)   )       up = 0;

    return 0.5 * (right.x - left.x + top.y - bottom.y + up.z - down.z);
}

// Jacobi computation
//-----------------------------------------

float PSComputeJacobi(GS_OUTPUT input) : SV_Target
{
    if(!IsWaterStrict(ToTexCoord(input.Tex)))
		return 0;

	float center = TexturePressure.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);
    
    float div = TextureTempScalar0.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);
    
    float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float left   = TexturePressure.SampleLevel(SamPointClamp, leftAddr, 0);
    float right  = TexturePressure.SampleLevel(SamPointClamp, rightAddr, 0);
    float bottom = TexturePressure.SampleLevel(SamPointClamp, bottomAddr, 0);
    float top    = TexturePressure.SampleLevel(SamPointClamp, topAddr, 0);
    float down   = TexturePressure.SampleLevel(SamPointClamp, downAddr, 0);
    float up     = TexturePressure.SampleLevel(SamPointClamp, upAddr, 0);
	
	if(IsObstacle(leftAddr))   left   = center;
    if(IsObstacle(rightAddr))  right  = center;
    if(IsObstacle(bottomAddr)) bottom = center;
    if(IsObstacle(topAddr))    top    = center;
    if(IsObstacle(downAddr))   down   = center;
    if(IsObstacle(upAddr))     up     = center;

	if(!IsWaterStrict(leftAddr))     left = 0;
    if(!IsWaterStrict(rightAddr))   right = 0;
    if(!IsWaterStrict(bottomAddr)) bottom = 0;
    if(!IsWaterStrict(topAddr))       top = 0;
    if(!IsWaterStrict(downAddr))     down = 0;
    if(!IsWaterStrict(upAddr))         up = 0;

    return (left + right + bottom + top + up + down - div) / 6.0f;
}

// Projection
//-----------------------------------------

float3 PSProject(GS_OUTPUT input) : SV_Target
{  
	if(!IsWaterStrict(ToTexCoord(input.Tex)))
		return 0;
	
	float center = TexturePressure.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);

    float3 leftAddr   = ToTexCoord(float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = ToTexCoord(float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = ToTexCoord(float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = ToTexCoord(float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = ToTexCoord(float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));
	
	float left   = TexturePressure.SampleLevel(SamPointClamp, leftAddr, 0);
    float right  = TexturePressure.SampleLevel(SamPointClamp, rightAddr, 0);
    float bottom = TexturePressure.SampleLevel(SamPointClamp, bottomAddr, 0);
    float top    = TexturePressure.SampleLevel(SamPointClamp, topAddr, 0);
    float down   = TexturePressure.SampleLevel(SamPointClamp, downAddr, 0);
    float up     = TexturePressure.SampleLevel(SamPointClamp, upAddr, 0);

	if(!IsWaterStrict(leftAddr))     left = 0;
    if(!IsWaterStrict(rightAddr))   right = 0;
    if(!IsWaterStrict(bottomAddr)) bottom = 0;
    if(!IsWaterStrict(topAddr))       top = 0;
    if(!IsWaterStrict(downAddr))     down = 0;
    if(!IsWaterStrict(upAddr))         up = 0;

    
    float3 mask = float3(1, 1, 1);
    
	if(IsObstacle(leftAddr))   
	{
		left = center;
		mask.x = 0;
	}

    if(IsObstacle(rightAddr)) 
	{
		right = center;
		mask.x = 0;
	}

    if(IsObstacle(bottomAddr)) 
	{
		bottom = center;
		mask.y = 0;
	}

    if(IsObstacle(topAddr))    
	{
		top = center;
		mask.y = 0;
	}

    if(IsObstacle(downAddr))   
	{
		down = center;
		mask.z = 0;
	}

    if(IsObstacle(upAddr))     
	{
		up = center;
		mask.z = 0;
	}

    return mask * (TextureVelocity1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0).xyz - (0.5 * float3(right - left, top - bottom, up - down)));
}

// Box drawing (fill)
//-----------------------------------------

float4 PSFillBox(GS_OUTPUT input) : SV_Target
{
    if(input.Tex.x > BoxMin.x &&
       input.Tex.y > BoxMin.y &&
       input.Tex.z > BoxMin.z &&
       input.Tex.x < BoxMax.x &&
       input.Tex.y < BoxMax.y &&
       input.Tex.z < BoxMax.z)
       return float4(1, 1, 1, -1);
     else
       return TextureColor0.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);
}

// Box drawing (border)
//-----------------------------------------

float PSBorderBox(GS_OUTPUT input) : SV_Target
{
    if((int)input.Tex.x == 0 ||
       (int)input.Tex.y == 0 ||
       (int)input.Tex.z == 0 ||
       (int)input.Tex.x == TextureWidth  - 1 ||
       (int)input.Tex.y == TextureHeight - 1 ||
       (int)input.Tex.z == TextureDepth  - 1)
       return 1.0f;
     else
       return -1.0f;
}

float4 PSDetectInterface(GS_OUTPUT input) : SV_Target
{
	float center = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0).w;

    float3 leftAddr   = (float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = (float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = (float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = (float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = (float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = (float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));

	float left   = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(leftAddr), 0).w;
    float right  = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(rightAddr), 0).w;
    float bottom = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(bottomAddr), 0).w;
    float top    = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(topAddr), 0).w;
    float down   = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(downAddr), 0).w;
    float up     = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(upAddr), 0).w;

	bool pos = (center > 0);
	float minDist = 999;
	float dist;
	float3 closestPoint;
	
	if(pos != (left > 0))
	{
		dist = center / (center - left);
	//	if(dist < minDist)		// 100% true
		{
			minDist = dist;
			closestPoint = leftAddr;
		}
	}

	if(pos != (right > 0))
	{
		dist = center / (center - right);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = rightAddr;
		}
	}

	if(pos != (bottom > 0))
	{
		dist = center / (center - bottom);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = bottomAddr;
		}
	}

	if(pos != (top > 0))
	{
		dist = center / (center - top);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = topAddr;
		}
	}

	if(pos != (down > 0))
	{
		dist = center / (center - down);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = downAddr;
		}
	}

	if(pos != (up > 0))
	{
		dist = center / (center - up);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = upAddr;
		}
	}
	
	if(minDist > 998)
		return float4(-1, -1, -1, 999 * sign(center));	// Don't know dist yet, set w = +/-999(infinity)
	else
		return float4(closestPoint, minDist * sign(center));	// Return signed distance 

}

float4 PSMarch(GS_OUTPUT input) : SV_Target
{
	float4 center = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);

	if(center.r > -0.5f)		// Dist is not @ infinity, save time
		return center;

    float3 leftAddr   = (float3(input.Tex.x - 1, input.Tex.y, input.Tex.z));
	float3 rightAddr  = (float3(input.Tex.x + 1, input.Tex.y, input.Tex.z));
	float3 bottomAddr = (float3(input.Tex.x, input.Tex.y - 1, input.Tex.z));
	float3 topAddr    = (float3(input.Tex.x, input.Tex.y + 1, input.Tex.z));
	float3 downAddr   = (float3(input.Tex.x, input.Tex.y, input.Tex.z - 1));
	float3 upAddr     = (float3(input.Tex.x, input.Tex.y, input.Tex.z + 1));

	float4 left   = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(leftAddr), 0);
    float4 right  = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(rightAddr), 0);
    float4 bottom = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(bottomAddr), 0);
    float4 top    = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(topAddr), 0);
    float4 down   = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(downAddr), 0);
    float4 up     = TextureColor1.SampleLevel(SamPointClamp, ToTexCoord(upAddr), 0);

	float minDist = 999;
	float dist;
	float3 closestPoint;

	if(left.r > -0.5f)
	{
		dist = length(left.xyz - input.Tex);
		//if(dist < minDist)	// 100% true
		{
			minDist = dist;
			closestPoint = left.xyz;
		}
	}

	if(right.r > -0.5f)
	{
		dist = length(right.xyz - input.Tex);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = right.xyz;
		}
	}

	if(bottom.r > -0.5f)
	{
		dist = length(bottom.xyz - input.Tex);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = bottom.xyz;
		}
	}

	if(top.r > -0.5f)
	{
		dist = length(top.xyz - input.Tex);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = top.xyz;
		}
	}

	if(down.r > -0.5f)
	{
		dist = length(down.xyz - input.Tex);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = down.xyz;
		}
	}

	if(up.r > -0.5f)
	{
		dist = length(up.xyz - input.Tex);
		if(dist < minDist)
		{
			minDist = dist;
			closestPoint = up.xyz;
		}
	}

	if(minDist > 998)
		return float4(-1, -1, -1, center.w);	// Return old dist
	else
		return float4(closestPoint, minDist * sign(center.w));	// Return signed distance

}

float4 PSExtrapolateVelocity(GS_OUTPUT input) : SV_Target
{	
	float3 pos = input.Tex;

	float4 levelSet = TextureColor0.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);
	if(levelSet.w < 0)
		return TextureVelocity0.SampleLevel(SamPointClamp, ToTexCoord(input.Tex), 0);
	else
		if(levelSet.x > -0.1f)
			return TextureVelocity0.SampleLevel(SamPointClamp, ToTexCoord(levelSet.xyz), 0);
		else
			return 0;
}

//-----------------------------------------

technique10 AdvectVelocity
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectVelocity()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 AdvectColor
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSAdvectColor()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 ComputeDivergence
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSComputeDivergence()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 ComputeJacobi
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSComputeJacobi()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 Project
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSProject()));
        SetRasterizerState(CullNone);
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 FillBox
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSFillBox()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 BorderBox
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSBorderBox()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 DetectInterface
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSDetectInterface()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 March
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSMarch()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 ExtrapolateVelocity
{
    pass P0
    {
        SetVertexShader(CompileShader(vs_4_0, VS()));
        SetGeometryShader(CompileShader(gs_4_0, GS()));
        SetPixelShader(CompileShader(ps_4_0, PSExtrapolateVelocity()));
        SetRasterizerState(CullNone);
        SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}