// Global objects
//-----------------------------------------
SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

BlendState AlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = ONE;
    DestBlendAlpha = ONE;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
}; 

BlendState NoAlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
    RenderTargetWriteMask[0] = 0x0F;
};


RasterizerState cullBack
{
	CullMode = Back;
};

RasterizerState cullFront
{
	CullMode = Front;
};

RasterizerState cullNone
{
	CullMode = None;
};

matrix Projection;
matrix View, ViewInv;
matrix World, WorldInv;
float3 camPos;

Texture3D txObject;
Texture3D txGradient;
Texture1D txTransfer;
Texture2D txFront;
Texture2D txBack;
TextureCube txSky;

float InvTextureWidth, InvTextureHeight, InvTextureDepth;
//-----------------------------------------




// Box rendering
//-----------------------------------------
struct PS_INPUT_BOX
{
    float4 Pos : SV_POSITION;
    float3 Tex : TEXCOORD0;
};

PS_INPUT_BOX VSBox(float4 pos : POSITION)
{
    PS_INPUT_BOX output;
								// Perspective transform, do not project object coords
    float4 posInView = mul(pos, mul(World, View));
    output.Pos = mul(posInView, Projection);
    
    output.Tex = pos.xyz;
    
    return output;
}

float4 PSBox(PS_INPUT_BOX input) : SV_Target
{
    return float4(input.Tex, 1.0f);
}
//-----------------------------------------

// Ray-casting
//-----------------------------------------

struct PS_INPUT_CAST
{
    float4 Pos : SV_POSITION;
    float3 Tex : TEXCOORD0;
};

PS_INPUT_CAST VSCast(float4 pos : POSITION)
{
    PS_INPUT_CAST output;
    output.Pos = pos;
    output.Tex = pos / 2 + 0.5;  // Convert [-1 1] to [0 1]
    output.Tex.y = 1.0f - output.Tex.y;

    return output;
}

float3 lightPos;
float steps, shadowSteps;
bool shading, shadows;

float3 CalcGradient(float3 pos)
{
	float3 leftAddr   = (float3(pos.x - InvTextureWidth, pos.y, pos.z));
	float3 rightAddr  = (float3(pos.x + InvTextureWidth, pos.y, pos.z));
	float3 bottomAddr = (float3(pos.x, pos.y - InvTextureHeight, pos.z));
	float3 topAddr    = (float3(pos.x, pos.y + InvTextureHeight, pos.z));
	float3 downAddr   = (float3(pos.x, pos.y, pos.z - InvTextureDepth));
	float3 upAddr     = (float3(pos.x, pos.y, pos.z + InvTextureDepth));

	float left   = txObject.SampleLevel(samLinear, leftAddr, 0).a;
    float right  = txObject.SampleLevel(samLinear, rightAddr, 0).a;
    float bottom = txObject.SampleLevel(samLinear, bottomAddr, 0).a;
    float top    = txObject.SampleLevel(samLinear, topAddr, 0).a;
    float down   = txObject.SampleLevel(samLinear, downAddr, 0).a;
    float up     = txObject.SampleLevel(samLinear, upAddr, 0).a;
	return normalize(float3(right - left, top - bottom, up - down));
}

float4 PSCast(PS_INPUT_CAST input) : SV_Target
{
    float4 vecFront;
    vecFront = txFront.Sample(samLinear, input.Tex.xy);

    float4 vecBack;
    vecBack = txBack.Sample(samLinear, input.Tex.xy);
    
    if(vecBack.a == 0.0f || vecFront.a == 0.0f)
		return float4(0, 0, 0, 0);
    
    float dz = 0.001;
    float3 vecDir = vecBack - vecFront;
    float vecLen = length(vecDir);
    int maxiter = int(floor(vecLen / dz));
    float3 vecDirNorm = vecDir  / vecLen;
	vecDir = vecDirNorm * dz;
	
	
    float3 v = vecFront;
	float4 color;

    for(int i = 0 ; i < 2000 ; i++)
    {
		v += vecDir;
		if(i >= maxiter)
			break;
				
		color = txObject.SampleLevel(samLinear, v, 0);
		if(color.a < 0)  // We're in the water
		{
			float3 normal = -CalcGradient(v);
			float3 lightPos = float3(-10, 10, -10);

			float3 light = normalize(lightPos - v);

			float3 reflectedView = 2.0 * dot(vecDirNorm, normal) * normal - vecDirNorm;
			
			color.a = 0.9f;

			// could be done better with snell's law
			color.rgb = 0.5 * txSky.SampleLevel(samLinear, vecDirNorm, 0).rgb + 0.5 * txSky.SampleLevel(samLinear, -reflectedView, 0).rgb;

			color.rgb *= min(0.8f + max(dot(normal, light), 0), 1);	// a bit of lighting too

			return color;
		}
	}
	
	return float4(1,1,1,0.2);	// Nothing found

}

//-----------------------------------------

PS_INPUT_BOX VSSkybox(float4 pos : POSITION)
{
    PS_INPUT_BOX output;
    
    pos.xyz = pos.xyz - 0.5f;
	pos.xyz *= 6;
	float4 posInWorld = mul(pos, World);		// Quick hack: transform skybox with 3d-tex world matrix to make it look as if the camera was rotating
    output.Pos = mul(posInWorld, Projection);
    
    output.Tex = pos.xyz;
    
    return output;
}

float4 PSSkybox(PS_INPUT_BOX input) : SV_Target
{
    return txSky.Sample(samLinear, input.Tex);
}

//-----------------------------------------


technique10 RenderBox
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSBox()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBox()));
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
    
    pass P1
    {
		SetRasterizerState(cullFront);
        SetVertexShader(CompileShader(vs_4_0, VSBox()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBox()));
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderCast
{
    pass P0
    {
        SetRasterizerState(cullBack);
        SetVertexShader(CompileShader(vs_4_0, VSCast()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSCast()));
		SetBlendState(AlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}

technique10 RenderSkybox
{
    pass P0
    {
        SetRasterizerState(cullNone);
        SetVertexShader(CompileShader(vs_4_0, VSSkybox()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSkybox()));
		SetBlendState(NoAlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
    }
}